import logging

import numpy as np
from scipy import constants

from haloreader.type_guards import is_ndarray
from haloreader.variable import Variable

log = logging.getLogger(__name__)


def compute_beta(
    intensity: Variable, range_: Variable, focus: Variable, wavelength: Variable
) -> Variable:
    # pylint: disable=invalid-name
    """
    Parameters
    ----------
    range_
        distance from the instrument
    focus
        focal length og the telescope for the transmitter and receiver
    lambda_
        laser wavelength

    Local variables
    ---------------
    eta
        detector quantum efficiency
    E
        beam energy
    nu
        optical frequency
    h
        planc's constant
    c
        speed of light
    B
        reveiver bandwidth

    References
    ----------
    Methodology for deriving the telescope focus function and
    its uncertainty for a heterodyne pulsed Doppler lidar
        authors:  Pyry Pentikäinen, Ewan James O'Connor,
            Antti Juhani Manninen, and Pablo Ortiz-Amezcua
        doi: https://doi.org/10.5194/amt-13-2849-2020
    """
    if not is_ndarray(intensity.data):
        raise TypeError
    if not is_ndarray(range_.data):
        raise TypeError
    if not isinstance(wavelength.data, float):
        raise TypeError
    if wavelength.units != "m":
        raise NotImplementedError(
            f'Expected wavelength units "m", got "{wavelength.units}".'
        )

    r = range_.data
    h = constants.Planck
    eta = 1
    c = constants.speed_of_light
    E = 1e-5
    lambda_ = wavelength.data
    nu = c / lambda_
    B = 5e7
    A_e = compute_effective_receiver_energy(range_, focus, lambda_)
    snr = intensity.data - 1
    # ref: https://doi.org/10.5194/amt-13-2849-2020
    beta = 2 * h * nu * B * r**2 * snr / (eta * c * E * A_e)
    return Variable(
        name="beta",
        long_name="attenuated backscatter coefficient",
        comment=(
            "Experimental variable. Computed using uncalibrated/placeholder values."
        ),
        units="m-1 sr-1",
        dimensions=intensity.dimensions,
        data=beta,
    )


def compute_effective_receiver_energy(
    range_: Variable, focus: Variable, lambda_: float
) -> np.ndarray:
    # pylint: disable=invalid-name
    """
    Parameters
    ----------
    range_
        distance from the instrument
    focus
        effective focal length of the telescope for the transmitter and receiver
    lambda_
        laser wavelength
    """
    if not is_ndarray(range_.data):
        raise TypeError
    log.warning(
        "Using placeholder values from https://doi.org/10.5194/amt-13-2849-2020"
    )
    r = range_.data
    D = 25e-3  # effective_diameter_of_gaussian_beam
    f = focus.data  # effective_focal_length
    return (
        np.pi
        * D**2
        / (4 * (1 + (np.pi * D**2 / (4 * lambda_ * r)) ** 2 * (1 - r / f) ** 2))
    )



import datetime
import logging
import pathlib
from io import BytesIO

import requests
import urllib3

from haloreader.halo import Halo, HaloBg
from haloreader.read import read, read_bg
from haloreader.scantype import ScanType

log = logging.getLogger(__name__)


class Session(requests.Session):
    def __init__(self) -> None:
        super().__init__()
        retries = urllib3.util.retry.Retry(total=10, backoff_factor=0.2)
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)
        self.mount("https://", adapter)
        self.url = "https://cloudnet.fmi.fi/api/raw-files"

    def get_metadata(
        self, site: str, date_from: datetime.date, date_to: datetime.date
    ) -> list:
        records = self.get(
            self.url,
            params={
                "instrument": "halo-doppler-lidar",
                "site": site,
                "dateFrom": str(date_from),
                "dateTo": str(date_to),
            },
        ).json()
        if not isinstance(records, list):
            raise TypeError
        return records


def get_halo_cloudnet(
    site: str, date: datetime.date, scantype: ScanType = ScanType.STARE
) -> tuple[Halo | None, HaloBg | None]:
    ses = Session()
    log.info("Fetching metadata for %s", date)
    records = ses.get_metadata(
        site, date_from=date - datetime.timedelta(days=60), date_to=date
    )
    bg_records = [r for r in records if HaloBg.is_bgfilename(r["filename"])]
    halo_records = [
        r
        for r in records
        if r["measurementDate"] == date.isoformat()
        and ScanType.from_filename(r["filename"]) == scantype
    ]
    halo_bytes = [_recor2bytes(r, ses) for r in halo_records]
    bg_bytes = [_recor2bytes(r, ses) for r in bg_records]
    bg_filenames = [r["filename"] for r in bg_records]
    return read(halo_bytes), read_bg(bg_bytes, bg_filenames)


def _recor2bytes(record: dict, session: Session) -> BytesIO:
    path = pathlib.Path("cache", record["uuid"])
    if path.exists():
        with path.open("rb") as f:
            file_io = BytesIO(f.read())
    else:
        log.info("Downloading and caching %s", record["filename"])
        file_bytes = session.get(record["downloadUrl"]).content
        path.parent.mkdir(exist_ok=True)
        with path.open("wb") as f:
            f.write(file_bytes)
        file_io = BytesIO(file_bytes)
    return file_io



import numpy as np
from scipy.ndimage import uniform_filter

from haloreader.type_guards import is_ndarray
from haloreader.variable import Variable


def compute_noise_screen(intensity: Variable, doppler_velocity: Variable) -> Variable:
    if not is_ndarray(intensity.data) or not is_ndarray(doppler_velocity.data):
        raise TypeError
    # kernel size and threshold values have been chosen just by
    # visually checking the output
    intensity_mean_mask = uniform_filter(intensity.data, size=(21, 3)) > 1.0025
    velocity_abs_mean_mask = (
        uniform_filter(np.abs(doppler_velocity.data), size=(21, 3)) < 2
    )
    bad_gates = np.zeros_like(intensity.data, dtype=bool)
    bad_gates[:, :3] = True
    below_one = intensity.data < 1
    return Variable(
        name="noise_screen",
        dimensions=intensity.dimensions,
        data=np.logical_not(intensity_mean_mask | velocity_abs_mean_mask)
        | bad_gates
        | below_one,
    )



import time
from datetime import datetime
from types import TracebackType

UNIX_TIME_UNIT = "seconds since 1970-01-01 00:00:00 +0000"
UNIX_TIME_FMT = "%Y-%m-%d %H:%M:%S %z"
CLOUDNET_TIME_UNIT_FMT = "hours since %Y-%m-%d %H:%M:%S %z"


class Timer:
    def __init__(self, name: str = "timer"):
        self.name = name
        self.start = 0.0

    def __enter__(self) -> None:
        self.start = time.perf_counter()

    def __exit__(
        self, exc_type: type, exc_value: AttributeError, exc_traceback: TracebackType
    ) -> None:
        end = time.perf_counter()
        runtime = end - self.start
        print(f"{self.name}: {runtime}")


def two_column_format(key: str, vals: list, left_width: int) -> str:
    if len(vals) == 0:
        return key
    str_ = key.ljust(left_width, " ") + str(vals[0])
    for v in vals[1:]:
        str_ += "\n" + " " * left_width + str(v)
    return str_


def indent_str(str_: str, width: int) -> str:
    indented_lines = []
    for line in str_.split("\n"):
        indented_lines.append(" " * width + line)
    return "\n".join(indented_lines)


def timestamp2str(stamp: float | list[float]) -> str | list[str]:
    def _timestamp2str(stamp: float) -> str:
        return datetime.utcfromtimestamp(stamp).strftime("%Y-%m-%d %H:%M:%S")

    if isinstance(stamp, list):
        return [_timestamp2str(s) for s in stamp]
    return _timestamp2str(stamp)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, TypeAlias, TypeGuard, runtime_checkable

import netCDF4
import numpy as np
from matplotlib.axes import Axes

from haloreader.exceptions import MergeError, NetCDFWriteError
from haloreader.type_guards import (
    is_fancy_index,
    is_float_list,
    is_int_list,
    is_ndarray,
    is_ndarray_list,
)

DataType: TypeAlias = np.ndarray | int | float | None


@dataclass(slots=True)
class Variable:
    name: str
    standard_name: str | None = None
    long_name: str | None = None
    comment: str | None = None
    calendar: str | None = None
    units: str | None = None
    dimensions: tuple[str, ...] | None = None
    data: DataType = None

    @classmethod
    def is_variable_list(cls, val: list[Any]) -> TypeGuard[list[Variable]]:
        return all(isinstance(x, Variable) for x in val)

    def nc_create_dimension(self, nc: netCDF4.Dataset) -> None:
        nc.createDimension(self.name, None)

    def nc_write(
        self,
        nc: netCDF4.Dataset,
        nc_map: dict[str, dict] | None = None,
        nc_exclude: dict[str, set] | None = None,
    ) -> None:
        nc_exclude_var = (
            nc_exclude.get("variables", set()) if nc_exclude is not None else set()
        )
        if self.name in nc_exclude_var:
            return
        nc_map_var = nc_map.get("variables", {}) if nc_map is not None else {}
        var_name = nc_map_var.get(self.name, self.name)
        nc_dtype = _choose_nc_dtype(self)
        dimensions = self.dimensions if self.dimensions is not None else ()
        mapped_dimensions = tuple((nc_map_var.get(dim, dim) for dim in dimensions))
        for dim in mapped_dimensions:
            if not _dimension_exists(nc, dim):
                nc.createDimension(dim, None)
        nc_var = nc.createVariable(var_name, nc_dtype, mapped_dimensions, zlib=True)
        nc_var[:] = self.data if self.data is not None else []
        for attr_name in set(self.__dataclass_fields__.keys()) - {
            "name",
            "dimensions",
            "data",
        }:
            attr_val = getattr(self, attr_name)
            if attr_val:
                setattr(nc_var, attr_name, attr_val)

    @classmethod
    def like(
        cls,
        var: Variable,
        data: DataType,
    ) -> Variable:
        return Variable(
            **{
                attr_name: getattr(var, attr_name)
                for attr_name in set(var.__dataclass_fields__.keys()) - {"data"}
            },
            data=data,
        )

    def take(self, index_list: list[int], axis: int = 0) -> Variable:
        if not is_ndarray(self.data):
            raise TypeError
        index = tuple(
            (index_list if axis == i else slice(None) for i in range(self.data.ndim))
        )
        if not is_fancy_index(index):
            raise TypeError
        return Variable.like(self, data=self.data[index])

    @classmethod
    def merge(cls, vars_: list[Variable]) -> Variable | None:
        _check_merge(vars_)

        if len(vars_) == 0:
            return None
        if len(vars_) == 1:
            return vars_[0]
        first_var = vars_[0]

        if first_var.name == "range":
            data: DataType = _reduce_range(vars_)
        elif first_var.dimensions is None:
            data = _get_scalar_data(vars_)
        else:
            data = _concatenate_along_first_dimension(vars_)
        return Variable(
            **{
                attr_name: getattr(first_var, attr_name)
                for attr_name in set(first_var.__dataclass_fields__.keys()) - {"data"}
            },
            data=data,
        )

    def plot(self, ax: Axes) -> None:
        if "intensity" in self.name:
            _plot_intensity(self, ax)
        elif "background" in self.name:
            _plot_background(self, ax)
        elif "doppler_velocity" in self.name:
            _plot_doppler_velocity(self, ax)
        elif "beta" in self.name:
            _plot_beta(self, ax)
        else:
            raise NotImplementedError(
                f"Plotting not implemented for variable {self.name}"
            )


@runtime_checkable
class VariableWithNumpyData(Protocol):
    data: np.ndarray
    name: str
    units: str
    dimensions: tuple[str, ...]


def _plot_intensity(var: Variable, ax: Axes) -> None:
    vdelta = 1e-3
    vmin, vmax = (1 - vdelta, 1 + vdelta)
    if not isinstance(var.data, np.ndarray):
        raise TypeError
    ax.imshow(
        var.data.T,
        origin="lower",
        vmin=vmin,
        vmax=vmax,
        aspect="auto",
        interpolation="none",
    )
    ax.set_title(var.name)


def _plot_beta(var: Variable, ax: Axes) -> None:
    if not isinstance(var.data, np.ndarray):
        raise TypeError
    vmin, vmax = (1e-7, 1e-4)
    ax.imshow(
        var.data.T,
        origin="lower",
        aspect="auto",
        interpolation="none",
        vmin=vmin,
        vmax=vmax,
    )
    ax.set_title(var.name)


def _plot_doppler_velocity(var: Variable, ax: Axes) -> None:
    if not isinstance(var.data, np.ndarray):
        raise TypeError
    vdelta = 4
    vmin, vmax = (-vdelta, vdelta)
    ax.imshow(
        var.data.T,
        origin="lower",
        aspect="auto",
        cmap="bwr",
        vmin=vmin,
        vmax=vmax,
        interpolation="none",
    )
    ax.set_title(var.name)


def _plot_background(var: Variable, ax: Axes) -> None:
    if not isinstance(var.data, np.ndarray):
        raise TypeError
    mean = var.data.mean()
    std = var.data.std()
    vmin, vmax = (mean - std, mean + std)
    ax.imshow(
        var.data.T,
        origin="lower",
        vmin=vmin,
        vmax=vmax,
        aspect="auto",
        interpolation="none",
    )
    ax.set_title(var.name)


def _dimension_exists(nc: netCDF4.Dataset | None, dim: str) -> bool:
    if nc is None:
        return False
    if dim in nc.dimensions:
        return True
    return _dimension_exists(nc.parent, dim)


def _choose_nc_dtype(var: Variable) -> str:
    if var.data is None:
        return "f4"
    if isinstance(var.data, float):
        return "f4"
    if isinstance(var.data, int):
        return "i4"
    if isinstance(var.data, np.ndarray):
        if var.name == "time" and var.data.dtype.kind == "f":
            return "f8"
        if var.data.dtype.kind == "f":
            return "f4"
        if var.data.dtype.kind == "i":
            return "i4"
        raise NetCDFWriteError
    raise NetCDFWriteError


def _check_merge(vars_: list[Variable]) -> None:
    if len(vars_) < 2:
        return
    first_var = vars_[0]
    for attr_name in set(first_var.__dataclass_fields__.keys()) - {"data"}:
        if not all(
            getattr(first_var, attr_name) == getattr(v, attr_name) for v in vars_[1:]
        ):
            raise MergeError


def _get_scalar_data(vars_: list[Variable]) -> float | int:
    scalars = [v.data for v in vars_]
    if not (is_float_list(scalars) or is_int_list(scalars)):
        raise TypeError
    if _scalars_equal(scalars):
        return scalars[0]
    raise MergeError("Cannot merge unequal dimensionless scalars")


def _scalars_equal(scalars: list[float] | list[int]) -> bool:
    if len(scalars) < 2:
        return True
    if is_int_list(scalars):
        return all(s == scalars[0] for s in scalars[1:])
    return all(np.isclose(s, scalars[0]) for s in scalars[1:])


def _reduce_range(vars_: list[Variable]) -> np.ndarray:
    range_list = [v.data for v in vars_]
    if not is_ndarray_list(range_list):
        raise TypeError
    if not _array_shapes_equal(range_list):
        raise MergeError("Cannot merge, range shapes unequal")
    if _arrays_close(range_list):
        return range_list[0]
    raise MergeError("Cannot merge unequal ranges")


def _concatenate_along_first_dimension(vars_: list[Variable]) -> np.ndarray:
    _check_concat(vars_)
    data_list = [v.data for v in vars_]
    if is_ndarray_list(data_list):
        return np.concatenate(data_list)
    raise TypeError


def _check_concat(vars_: list[Variable]) -> None:
    if len(vars_) < 2:
        return
    first_var = vars_[0]
    if not isinstance(first_var.data, np.ndarray):
        raise TypeError
    if isinstance(first_var.dimensions, tuple):
        ndim = len(first_var.dimensions)
    else:
        raise TypeError
    data_list = [v.data for v in vars_]
    if not is_ndarray_list(data_list):
        raise MergeError
    if ndim < 1:
        raise MergeError
    if not all(ndim == np.ndim(data) for data in data_list):
        raise MergeError
    first_shape = first_var.data.shape
    if ndim > 1 and not all(first_shape[1:] == d.shape[1:] for d in data_list):
        raise MergeError


def _arrays_close(array_list: list[np.ndarray]) -> bool:
    if len(array_list) < 2:
        return True
    first_arr = array_list[0]
    return all(np.allclose(first_arr, arr) for arr in array_list)


def _array_shapes_equal(array_list: list[np.ndarray]) -> bool:
    if len(array_list) < 2:
        return True
    first_arr = array_list[0]
    return all(first_arr.shape == arr.shape for arr in array_list)



import numpy as np

from haloreader.variable import Variable

from libc.stdlib cimport atof
from libc.string cimport strlen, strspn, strtok


def read_background(data_py: bytes) -> Variable:
    if b"\n" not in data_py:
        return _read_background_without_newlines(data_py)
    cdef char * data_c = data_py
    cdef char * token

    cdef unsigned long ntokens = 0
    token = strtok(data_c, " \r\n")
    while token != NULL:
        token = strtok(NULL, " \r\n")
        ntokens += 1

    data = np.zeros((1,ntokens), dtype=np.dtype("float"))
    cdef double [:,:] data_view = data

    cdef unsigned long i
    token = data_c
    for i in range(ntokens):
        data_view[0,i] = atof(token)
        token += strlen(token) + 1
        token += strspn(token, " \r\n")

    return Variable(
            name = "background",
            data = data,
            dimensions = ("time", "range"),
            )


def _read_background_without_newlines(data_py: bytes) -> Variable:
    cdef char * data_c = data_py
    cdef char ch
    cdef char[32] num

    cdef unsigned long i = 0
    cdef unsigned long k = 0
    cdef unsigned long j = 0
    cdef unsigned long nchars = len(data_py)

    cdef unsigned long ntokens = 0
    while k < nchars:
        if data_c[k] == 46: # if '.'
            ntokens += 1
        k += 1

    data = np.zeros((1,ntokens), dtype=np.dtype("float"))
    cdef double [:,:] data_view = data

    k = 0
    while k < nchars:
        if data_c[k] == 46: # '.' is 46
            for _ in range(7):
                num[j] = data_c[k]
                k += 1
                j += 1
            num[j] = b"\0"
            data[0,i] = atof(num)
            i += 1
            j = 0
        else:
            num[j] = data_c[k]
            k += 1
            j += 1
    return Variable(
            name = "background",
            data = data,
            dimensions = ("time", "range"),
            )



import argparse
import logging
import pathlib

import flask

from . import DEFAULT_ROOT, IMAGE_EXT

logging.basicConfig(level=logging.INFO)


parser = argparse.ArgumentParser("haloboard")
parser.add_argument("--dir", type=pathlib.Path, default=pathlib.Path(DEFAULT_ROOT))
args = parser.parse_args()
app = flask.Flask(
    __name__,
    template_folder=str(pathlib.Path(__file__).parent.joinpath("templates")),
)
ROOT = args.dir


def _find_images() -> list:
    img_dir = pathlib.Path(ROOT)
    return sorted(img_dir.glob(f"*.{IMAGE_EXT}"))


@app.route("/")
def index() -> str:
    image_paths = _find_images()
    image_paths = ["/" + str(p) for p in image_paths]
    return flask.render_template("index.html", image_paths=image_paths)


@app.route(f"/{ROOT}/<path:name>")
def serve_files(name: str) -> flask.wrappers.Response:
    return flask.send_from_directory(
        pathlib.Path(ROOT).resolve(), name, as_attachment=True
    )


@app.route("/static/<path:name>")
def serve_static(name: str) -> flask.wrappers.Response:
    return flask.send_from_directory(pathlib.Path(app.root_path, "static"), name)


def run() -> None:
    app.run(debug=True)


if __name__ == "__main__":
    run()



from __future__ import annotations

import datetime
import logging
import re
from dataclasses import dataclass
from typing import Any, Protocol, TypeGuard, runtime_checkable

import netCDF4
import numpy as np

import haloreader.attenuated_backscatter_coefficient
import haloreader.background_correction
import haloreader.screen
from haloreader.metadata import Metadata
from haloreader.type_guards import is_fancy_index, is_ndarray, is_none_list
from haloreader.utils import CLOUDNET_TIME_UNIT_FMT, UNIX_TIME_FMT, UNIX_TIME_UNIT
from haloreader.variable import Variable, VariableWithNumpyData

log = logging.getLogger(__name__)


@dataclass(slots=True)
class Halo:
    metadata: Metadata
    time: Variable
    range: Variable
    azimuth: Variable
    elevation: Variable
    doppler_velocity: Variable
    intensity_raw: Variable
    beta_raw: Variable
    spectral_width: Variable | None = None
    intensity: Variable | None = None
    beta: Variable | None = None
    beta_screened: Variable | None = None
    doppler_velocity_screened: Variable | None = None
    pitch: Variable | None = None
    roll: Variable | None = None

    def to_nc(
        self,
        nc_map: dict[str, dict] | None = None,
        nc_exclude: dict[str, set] | None = None,
    ) -> memoryview:
        nc = netCDF4.Dataset("inmemory.nc", "w", memory=1028)
        for attr_name in self.__dataclass_fields__.keys():
            halo_attr = getattr(self, attr_name)
            if halo_attr is not None:
                halo_attr.nc_write(nc, nc_map=nc_map, nc_exclude=nc_exclude)
        nc_buf = nc.close()
        if isinstance(nc_buf, memoryview):
            return nc_buf
        raise TypeError

    @classmethod
    def merge(cls, halos: list[Halo]) -> Halo | None:
        if len(halos) == 0:
            return None
        if len(halos) == 1:
            return halos[0]
        halo_attrs: dict[str, Any] = {}
        sorted_halos = _sorted_halo_list(halos)

        for attr_name in cls.__dataclass_fields__.keys():
            halo_attr_list = [getattr(h, attr_name) for h in sorted_halos]
            if Metadata.is_metadata_list(halo_attr_list):
                halo_attrs[attr_name] = Metadata.merge(halo_attr_list)
            elif Variable.is_variable_list(halo_attr_list):
                halo_attrs[attr_name] = Variable.merge(halo_attr_list)
            elif is_none_list(halo_attr_list):
                halo_attrs[attr_name] = None
            else:
                raise TypeError

        halo = Halo(**halo_attrs)
        if not isinstance(halo.time.data, np.ndarray):
            raise TypeError
        if not _is_increasing(halo.time.data):
            halo.remove_profiles_with_duplicate_time()
        if not _is_increasing(halo.time.data):
            halo.remove_profiles_with_non_increasing_time()
        if not _is_increasing(halo.time.data):
            raise ValueError("Time must be increasing")
        return halo

    def remove_profiles_with_duplicate_time(self) -> None:
        if not is_ndarray(self.time.data):
            raise TypeError
        mask = _duplicate_time_mask(self.time.data)
        for attr_name in self.__dataclass_fields__.keys():
            halo_attr = getattr(self, attr_name)
            if (
                isinstance(halo_attr, Variable)
                and is_ndarray(halo_attr.data)
                and isinstance(halo_attr.dimensions, tuple)
                and self.time.name in halo_attr.dimensions
            ):
                index = tuple(
                    (
                        mask if d == self.time.name else slice(None)
                        for i, d in enumerate(halo_attr.dimensions)
                    )
                )
                if not is_fancy_index(index):
                    raise TypeError
                halo_attr.data = halo_attr.data[index]

    def remove_profiles_with_non_increasing_time(self) -> None:
        if not is_ndarray(self.time.data):
            raise TypeError
        mask = _non_increasing_time_mask(self.time.data)
        for attr_name in self.__dataclass_fields__.keys():
            halo_attr = getattr(self, attr_name)
            if (
                isinstance(halo_attr, Variable)
                and is_ndarray(halo_attr.data)
                and isinstance(halo_attr.dimensions, tuple)
                and self.time.name in halo_attr.dimensions
            ):
                index = tuple(
                    (
                        mask if d == self.time.name else slice(None)
                        for i, d in enumerate(halo_attr.dimensions)
                    )
                )
                if not is_fancy_index(index):
                    raise TypeError
                halo_attr.data = halo_attr.data[index]

    def convert_time_unit2cloudnet_time(self) -> None:
        _convert_timevar_unit2cloudnet_time(self.time)
        _convert_timevar_unit2cloudnet_time(self.metadata.start_time)

    def correct_background(self, halobg: HaloBg) -> None:
        if not is_ndarray(self.range.data):
            raise TypeError
        halobg_sliced = halobg.slice_range(len(self.range.data))
        p_amp = halobg_sliced.amplifier_noise()
        intensity_step1 = (
            haloreader.background_correction.background_measurement_correction(
                self.time,
                self.intensity_raw,
                halobg_sliced.time,
                halobg_sliced.background,
                p_amp,
            )
        )
        signalmask = haloreader.background_correction.threshold_signalmask(
            intensity_step1
        )
        self.intensity = haloreader.background_correction.snr_correction(
            intensity_step1, signalmask
        )

    def compute_beta(self) -> None:
        if not isinstance(self.intensity, Variable):
            raise TypeError
        log.warning("beta is computed using placeholder values")
        self.beta = haloreader.attenuated_backscatter_coefficient.compute_beta(
            self.intensity,
            self.range,
            self.metadata.focus_range,
            self.metadata.wavelength,
        )

    def compute_noise_screen(self) -> Variable:
        if not isinstance(self.intensity, Variable):
            raise TypeError
        return haloreader.screen.compute_noise_screen(
            self.intensity, self.doppler_velocity
        )

    def compute_beta_screened(self, screen: Variable) -> None:
        if not isinstance(self.beta, Variable):
            raise TypeError
        self.beta_screened = Variable(
            name="beta_screened",
            long_name="screened attenuated backscatter coefficient",
            comment=(
                "Experimental variable. Computed using uncalibrated/placeholder values."
            ),
            units="m-1 sr-1",
            dimensions=self.beta.dimensions,
            data=np.ma.masked_array(self.beta.data, mask=screen.data),
        )

    def compute_doppler_velocity_screened(self, screen: Variable) -> None:
        if not isinstance(self.beta, Variable):
            raise TypeError
        self.doppler_velocity_screened = Variable(
            name="doppler_velocity_screened",
            long_name="screened radial velocity (positive away from lidar)",
            units=self.doppler_velocity.units,
            dimensions=self.doppler_velocity.dimensions,
            data=np.ma.masked_array(self.doppler_velocity.data, mask=screen.data),
        )


def _convert_timevar_unit2cloudnet_time(var: Variable) -> None:
    """Converts time variable to cloudnet format.

    Format is: hours since [beginning of the day of the first
    measurement]
    """
    if not (is_ndarray(var.data) and len(var.data) > 0):
        return
    if not isinstance(var.units, str):
        raise TypeError
    if match_ := re.match(r"(seconds|hours) since (.*)", var.units):
        unit = match_.group(1)
        if not isinstance(unit, str):
            raise TypeError
        base = datetime.datetime.strptime(match_.group(2), UNIX_TIME_FMT)
        match unit:
            case "seconds" | "minutes" | "hours" | "weeks":
                tdelta = datetime.timedelta(**{unit: var.data[0]})  # type: ignore
            case other_unit:
                raise ValueError(f"Unexpected unit {other_unit}")
        new_base = (base + tdelta).replace(hour=0, minute=0, second=0, microsecond=0)
        base_diff = (base - new_base).total_seconds()
        scale = datetime.timedelta(**{unit: 1}).total_seconds()  # type: ignore
        new_time_in_hours = np.array([(base_diff + scale * t) / 3600 for t in var.data])
        new_unit = new_base.strftime(CLOUDNET_TIME_UNIT_FMT)
        var.data = new_time_in_hours
        var.units = new_unit
    else:
        raise NotImplementedError


def _sorted_halo_list_key(halo: Halo) -> float:
    if not is_ndarray(halo.metadata.start_time.data):
        raise TypeError
    if halo.metadata.start_time.units != UNIX_TIME_UNIT:
        raise ValueError(f"Unexpected time units: {halo.metadata.start_time.units}")
    if len(halo.metadata.start_time.data) != 1:
        raise ValueError
    start_time = halo.metadata.start_time.data[0]
    if not isinstance(start_time, float):
        raise TypeError
    return start_time


def _sorted_halo_list(halos: list[Halo]) -> list[Halo]:
    return sorted(halos, key=_sorted_halo_list_key)


@dataclass(slots=True)
class HaloBg:
    time: Variable
    range: Variable
    background: Variable

    def to_nc(self) -> memoryview:
        nc = netCDF4.Dataset("inmemory.nc", "w", memory=1028)
        self.time.nc_create_dimension(nc)
        self.range.nc_create_dimension(nc)
        for attr_name in self.__dataclass_fields__.keys():
            halobg_attr = getattr(self, attr_name)
            if halobg_attr is not None:
                halobg_attr.nc_write(nc)
        nc_buf = nc.close()
        if isinstance(nc_buf, memoryview):
            return nc_buf
        raise TypeError

    def slice_range(self, slice_: int | slice) -> HaloBg:
        if isinstance(slice_, int):
            slice_ = slice(slice_)
        if not is_ndarray(self.time.data):
            raise TypeError
        if not is_ndarray(self.range.data):
            raise TypeError
        if not is_ndarray(self.background.data):
            raise TypeError
        return HaloBg(
            time=Variable.like(self.time, data=self.time.data),
            range=Variable.like(self.range, data=self.range.data[slice_]),
            background=Variable.like(
                self.background, data=self.background.data[:, slice_]
            ),
        )

    @classmethod
    def is_halobg_with_numpy_data(
        cls, halobg: HaloBg
    ) -> TypeGuard[HaloBgWithNumpyData]:
        return isinstance(halobg, HaloBgWithNumpyData)

    @classmethod
    def is_list_of_halobgs(cls, halobgs: list[Any]) -> TypeGuard[list[HaloBg]]:
        return all(isinstance(hbg, HaloBg) for hbg in halobgs)

    @classmethod
    def is_list_of_halobgs_with_numpy_data(
        cls, halobgs: list[Any]
    ) -> TypeGuard[list[HaloBgWithNumpyData]]:
        return all(isinstance(hbg, HaloBgWithNumpyData) for hbg in halobgs)

    @classmethod
    def merge(cls, halobgs: list[HaloBg]) -> HaloBg | None:
        if len(halobgs) == 0:
            return None
        if len(halobgs) == 1:
            return halobgs[0]

        if not HaloBg.is_list_of_halobgs_with_numpy_data(halobgs):
            raise TypeError
        nranges_med = np.median([hbg.range.data.shape[0] for hbg in halobgs]).astype(
            int
        )
        halobgs_filtered = [
            hbg for hbg in halobgs if hbg.range.data.shape[0] == nranges_med
        ]
        if not HaloBg.is_list_of_halobgs(halobgs_filtered):
            raise TypeError
        nignored = len(halobgs) - len(halobgs_filtered)
        if nignored > 0:
            log.warning(
                "Ignoring %d/%d background profiles (mismatching # of range gates)",
                nignored,
                len(halobgs),
            )
        halobg_attrs: dict[str, Any] = {}
        sorted_halobgs = _sorted_halobg_list(halobgs_filtered)
        for attr_name in cls.__dataclass_fields__.keys():
            halobg_attr_list = [getattr(h, attr_name) for h in sorted_halobgs]
            if Variable.is_variable_list(halobg_attr_list):
                halobg_attrs[attr_name] = Variable.merge(halobg_attr_list)
            elif is_none_list(halobg_attr_list):
                halobg_attrs[attr_name] = None
            else:
                raise TypeError
        halobg = HaloBg(**halobg_attrs)
        if not isinstance(halobg.time.data, np.ndarray):
            raise TypeError
        if not _is_increasing(halobg.time.data):
            raise ValueError("Time must be increasing")
        return halobg

    @classmethod
    def is_bgfilename(cls, filename: str) -> bool:
        return filename.lower().startswith("background_")

    def amplifier_noise(self) -> Variable:
        if not HaloBg.is_halobg_with_numpy_data(self):
            raise TypeError
        _sum_over_gates = self.background.data.sum(axis=1)
        _normalised_bg = self.background.data / _sum_over_gates[:, np.newaxis]
        return Variable(
            name="p_amp",
            long_name=(
                "Mean of normalised background over time. "
                "Profiles are normalised with sum over gates"
            ),
            dimensions=("range",),
            data=_normalised_bg.mean(axis=0),
        )


@runtime_checkable
class HaloBgWithNumpyData(Protocol):
    time: VariableWithNumpyData
    range: VariableWithNumpyData
    background: VariableWithNumpyData


def _sorted_halobg_list_key(halobg: HaloBg) -> float:
    if not is_ndarray(halobg.time.data):
        raise TypeError
    if halobg.time.units != UNIX_TIME_UNIT:
        raise ValueError
    if len(halobg.time.data) != 1:
        raise ValueError
    time = halobg.time.data[0]
    if not isinstance(time, float):
        raise TypeError
    return time


def _sorted_halobg_list(halobgs: list[HaloBg]) -> list[HaloBg]:
    return sorted(halobgs, key=_sorted_halobg_list_key)


def _is_increasing(time: np.ndarray) -> bool:
    return bool(np.all(np.diff(time) > 0))


def _duplicate_time_mask(time: np.ndarray) -> np.ndarray:
    _mask = np.isclose(np.diff(time), 0)
    nremoved = _mask.sum()
    if nremoved > 0:
        log.debug("Removed %d/%d profiles (duplicate timestamps)", nremoved, len(time))
    return np.logical_not(np.insert(_mask, 0, False))


def _non_increasing_time_mask(time: np.ndarray) -> np.ndarray:
    if len(time) == 0:
        return np.zeros_like(time, dtype=bool)
    _mask = np.zeros_like(time, dtype=bool)
    largest_time = time[0]
    for i, current_time in enumerate(time[1:], start=1):
        if current_time <= largest_time:
            _mask[i] = True
        else:
            largest_time = current_time
    nremoved = _mask.sum()
    if nremoved > 0:
        log.debug(
            "Removed %d/%d profiles (non-increasing timestamps)", nremoved, len(time)
        )
    mask = np.logical_not(_mask)
    if not is_ndarray(mask):
        raise TypeError
    return mask



import datetime
from pathlib import Path

import numpy as np

from haloreader.read import read_bg

raw_files_pass = Path("tests/raw-files/pass/")


def timestr(unix_time):
    return datetime.datetime.utcfromtimestamp(unix_time).strftime("%Y-%m-%d %H:%M:%S")


def test_eriswil_background():
    bg_dir = raw_files_pass.joinpath("eriswil-2022-12-14-background")
    bg_00 = bg_dir.joinpath("Background_141222-000013.txt")
    bg_01 = bg_dir.joinpath("Background_141222-010013.txt")
    halobg = read_bg([bg_00, bg_01])

    times = halobg.time.data
    assert timestr(times[0]) == "2022-12-14 00:00:13"
    assert timestr(times[1]) == "2022-12-14 01:00:13"
    bg = halobg.background
    assert bg.dimensions == ("time", "range")
    assert bg.data.shape == (2, 250)
    assert np.isclose(bg.data[0, 0], 610890.0)
    assert np.isclose(bg.data[0, 1], 14318556.375)
    assert np.isclose(bg.data[-1, -2], 16885681.125)
    assert np.isclose(bg.data[-1, -1], 16881329.375)


def test_hyytiala_background():
    bg_dir = raw_files_pass.joinpath("hyytiala-2023-08-15-background")
    bg_00 = bg_dir.joinpath("Background_150823-122811.txt")
    halobg = read_bg([bg_00])

    times = halobg.time.data
    assert timestr(times[0]) == "2023-08-15 12:28:11"
    bg = halobg.background
    assert bg.dimensions == ("time", "range")
    assert bg.data.shape == (1, 400)
    assert np.isclose(bg.data[0, 0], 575587.333333)
    assert np.isclose(bg.data[0, 1], 14902110.166667)
